"""
This module contains examples of interest rate swaps.
These methods creates a timetable with floating point for time, which is now being deprecated.
See qablet_contracts/ir/swap to create timetables with timestamps.
"""

from typing import Dict, List

import numpy as np

from qablet_contracts.timetable import timetable_from_dicts


def simple_swap_period(
    ccy: str, start: float, end: float, fixed_rate: float, track: str = ""
) -> list:
    """Simple representation of a swap period, paying fixed, receiving floating rate.

    Args:
        ccy: the currency of the swap.
        start: the start of the period.
        end: the end of the period.
        fixed_rate: the fixed annual rate of the swap.
        track: an optional identifier for the contract.
    """
    return [
        {
            "track": track,
            "time": start,
            "op": "+",
            "quantity": 1,
            "unit": ccy,
        },
        {
            "track": track,
            "time": end,
            "op": "+",
            "quantity": -1 - fixed_rate * (end - start),
            "unit": ccy,
        },
    ]


def swap_timetable(
    ccy: str, times: List[float], strike_rate: float, track: str = ""
) -> Dict:
    """In a **Vanilla Swap**, at the end of each period the holder pays a fixed rate and receives a floating rate.
    In this simple version the floating rate payment is replaced by receiving notional at the beginning of the period
    and paying the notional at the end of the period.

    Args:
        ccy: the currency of the swap.
        times: the period times of the swap, including the inception and maturity.
        strike_rate: the strike rate of the swaption (in units, i.e. 0.02 means 200 bps).
        track: an optional identifier for the contract.

    Examples:
        >>> tt = swap_timetable("USD", [0.5, 1.0, 1.5], 0.05)
        >>> tt["events"].to_pandas()
          track  time op  quantity  unit
        1  .swp   0.5  +     1.000   USD
        2  .swp   1.0  +    -1.025   USD
        3  .swp   1.0  +     1.000   USD
        4  .swp   1.5  +    -1.025   USD
    """
    events = []
    # payment events
    for start, end in zip(times[0:-1], times[1:]):
        events.extend(
            simple_swap_period(ccy, start, end, strike_rate, track + ".swp")
        )

    return timetable_from_dicts(events)


def swaption_timetable(
    ccy: str, times: List[float], strike_rate: float, track: str = ""
) -> Dict:
    """In a **Vanilla swaption** the holder gets the opportunity to enter into the swap
    at the beginning of the first period.

    Args:
        ccy: the currency of the swap.
        times: the period times of the underlying swap, including the inception and maturity.
        strike_rate: the strike rate of the swaption (in units, i.e. 0.02 means 200 bps).
        track: an optional identifier for the contract.

    Examples:
        >>> tt = swaption_timetable("USD", [0.5, 1.0, 1.5], 0.05)
        >>> tt["events"].to_pandas()
          track  time op  quantity  unit
        0  .opt   0.5  >     1.000  .swp
        1  .swp   0.5  +     1.000   USD
        2  .swp   1.0  +    -1.025   USD
        3  .swp   1.0  +     1.000   USD
        4  .swp   1.5  +    -1.025   USD
    """

    # option expiration event at beginning of the swap
    events = [
        {
            "track": track + ".opt",
            "time": times[0],
            "op": ">",
            "quantity": 1,
            "unit": track + ".swp",
        }
    ]
    # payment events for the underlying swap
    for start, end in zip(times[0:-1], times[1:]):
        events.extend(
            simple_swap_period(ccy, start, end, strike_rate, track + ".swp")
        )

    return timetable_from_dicts(events)


def bermuda_swaption_timetable(
    ccy: str, times: List[float], strike_rate: float, track: str = ""
) -> Dict:
    """In a **Co-terminal Bermuda Swaption**, the holder can exercise his option at the beginning of each swap period.
    If exercised, the holder pays and receives all remaining payments of the swap. If not exercised, there are
    no payments in the next swap period. Irrespective of the time of exercise, the swap terminates at the same date.


    Args:
        ccy: the currency of the swap.
        times: the period times of the underlying swap, including the inception and maturity.
        strike_rate: the strike rate of the swaption (in units, i.e. 0.02 means 200 bps).
        track: an optional identifier for the contract.

    Examples:
        >>> tt = bermuda_swaption_timetable("USD", [0.5, 1.0, 1.5], 0.05)
        >>> tt["events"].to_pandas()
          track  time op  quantity  unit
        0  .opt   0.5  >     1.000  .swp
        1  .swp   0.5  +     1.000   USD
        2  .swp   1.0  +    -1.025   USD
        3  .opt   1.0  >     1.000  .swp
        4  .swp   1.0  +     1.000   USD
        5  .swp   1.5  +    -1.025   USD
    """

    events = []
    for start, end in zip(times[0:-1], times[1:]):
        # option expiration event before each period
        events.append(
            {
                "track": track + ".opt",
                "time": start,
                "op": ">",
                "quantity": 1,
                "unit": track + ".swp",
            }
        )
        # payment event for the underlying swap
        events.extend(
            simple_swap_period(ccy, start, end, strike_rate, track + ".swp")
        )

    return timetable_from_dicts(events)


if __name__ == "__main__":
    # Create the bermuda swaption
    times = np.linspace(
        1, 6, 6
    ).tolist()  # Start at 1 year, mature at 6 years.
    strike_rate = 0.03
    timetable = bermuda_swaption_timetable("USD", times, strike_rate)

    print(timetable["events"].to_pandas())



"""
This module contains examples of interest rate swaptions.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List

import pandas as pd

from qablet_contracts.ir.swap import simple_swap_period
from qablet_contracts.timetable import EventsMixin


@dataclass
class Swaption(EventsMixin):
    """A **Vanilla Swaption**.
    In a Vanilla swaption the holder gets the opportunity to enter into the swap at the beginning of the first period.

    Args:
        ccy: the currency of the swap.
        dates: the period datetimes of the underlying swap, including the inception and maturity.
        strike_rate: the strike rate of the swaption (in units, i.e. 0.02 means 200 bps).
        track: an optional identifier for the contract.

    Examples:
        >>> dates = pd.bdate_range(datetime(2023, 12, 31), datetime(2024, 12, 31), freq="2QE")
        >>> tt = Swap("USD", dates, strike_rate = 0.03).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time op  quantity  unit
        0  .opt 2023-12-31 00:00:00+00:00  >     1.000  .swp
        1  .swp 2023-12-31 00:00:00+00:00  +     1.000   USD
        2  .swp 2024-06-30 00:00:00+00:00  +    -1.015   USD
        3  .swp 2024-06-30 00:00:00+00:00  +     1.000   USD
        4  .swp 2024-12-31 00:00:00+00:00  +    -1.015   USD
    """

    ccy: str
    dates: List[datetime]
    strike_rate: float
    track: str = ""

    def events(self):
        # option expiration event at beginning of the swap
        events = [
            {
                "track": self.track + ".opt",
                "time": self.dates[0],
                "op": ">",
                "quantity": 1,
                "unit": self.track + ".swp",
            }
        ]
        # payment events for the underlying swap
        for start, end in zip(self.dates[0:-1], self.dates[1:]):
            events.extend(
                simple_swap_period(
                    self.ccy, start, end, self.strike_rate, self.track + ".swp"
                )
            )

        return events


@dataclass
class BermudaSwaption(EventsMixin):
    """In a **Co-terminal Bermuda Swaption**, the holder can exercise his option at the beginning of each swap period.
    If exercised, the holder pays and receives all remaining payments of the swap. If not exercised, there are
    no payments in the next swap period. Irrespective of the time of exercise, the swap terminates at the same date.

    Args:
        ccy: the currency of the swap.
        dates: the period datetimes of the underlying swap, including the inception and maturity.
        strike_rate: the strike rate of the swaption (in units, i.e. 0.02 means 200 bps).
        track: an optional identifier for the contract.

    Examples:
        >>> dates = pd.bdate_range(datetime(2023, 12, 31), datetime(2024, 12, 31), freq="2QE")
        >>> tt = Swap("USD", dates, strike_rate = 0.03).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time op  quantity  unit
        0  .opt 2023-12-31 00:00:00+00:00  >     1.000  .swp
        1  .swp 2023-12-31 00:00:00+00:00  +     1.000   USD
        2  .swp 2024-06-30 00:00:00+00:00  +    -1.015   USD
        3  .opt 2024-06-30 00:00:00+00:00  >     1.000  .swp
        4  .swp 2024-06-30 00:00:00+00:00  +     1.000   USD
        5  .swp 2024-12-31 00:00:00+00:00  +    -1.015   USD
    """

    ccy: str
    dates: List[datetime]
    strike_rate: float
    track: str = ""

    def events(self):
        events = []
        for start, end in zip(self.dates[0:-1], self.dates[1:]):
            # option expiration event before each period
            events.append(
                {
                    "track": self.track + ".opt",
                    "time": start,
                    "op": ">",
                    "quantity": 1,
                    "unit": self.track + ".swp",
                }
            )
            # payment event for the underlying swap
            events.extend(
                simple_swap_period(
                    self.ccy, start, end, self.strike_rate, self.track + ".swp"
                )
            )

        return events


if __name__ == "__main__":
    dates = pd.bdate_range(
        datetime(2023, 12, 31),
        datetime(2024, 12, 31),
        freq="2QE",
    )
    strike_rate = 0.03

    timetable = Swaption("USD", dates, strike_rate).timetable()
    print(timetable["events"].to_pandas())

    timetable = BermudaSwaption("USD", dates, strike_rate).timetable()
    print(timetable["events"].to_pandas())



"""
This module contains examples of creating timetables for equity rainbow options.
These methods creates a timetable with floating point for time, which is now being deprecated.
See qablet_contracts/eq/rainbow to create timetables with timestamps.
"""

from typing import List

from qablet_contracts.timetable import timetable_from_dicts


def rainbow_timetable(
    ccy: str,
    asset_names: List[str],
    strikes: List[float],
    notional: float,
    maturity: float,
    is_call: bool,
    track: str = "",
) -> dict:
    """Create timetable for an **Equity Rainbow Option**.

    Args:
        ccy: the currency of the option.
        asset_names: the name of the underlying assets.
        strikes: the option strikes.
        notional: the notional of the option.
        maturity: the maturity of the option in years.
        is_call: true if the option is a call.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = rainbow_timetable(
        >>>     "USD",
        >>>     ["SPX", "FTSE", "N225"],
        >>>     [5087, 7684, 39100],
        >>>     100_000,
        >>>     0.5,
        >>>     True
        >>> )
        >>> tt["events"].to_pandas()  # Call
          track  time op       quantity  unit
        0         0.5  + -100000.000000   USD
        1         0.5  >      19.657952   SPX
        2         0.5  >      13.014055  FTSE
        3         0.5  >       2.557545  N225
        4         0.5  +  100000.000000   USD

        >>> tt = rainbow_timetable(
        >>>     "USD",
        >>>     ["SPX", "FTSE", "N225"],
        >>>     [5087, 7684, 39100],
        >>>     100_000,
        >>>     0.5,
        >>>     False
        >>> )
        >>> tt["events"].to_pandas()  # Put
          track  time op       quantity  unit
        0         0.5  +  100000.000000   USD
        1         0.5  >     -19.657952   SPX
        2         0.5  >     -13.014055  FTSE
        3         0.5  >      -2.557545  N225
        4         0.5  + -100000.000000   USD
    """

    sign = 1 if is_call else -1
    # Pay the initial strike
    events = [
        {
            "track": "",
            "time": maturity,
            "op": "+",
            "quantity": -notional * sign,
            "unit": ccy,
        },
    ]

    # Options to receive any of the assets
    for asset, strike in zip(asset_names, strikes):
        events.append(
            {
                "track": "",
                "time": maturity,
                "op": ">",
                "quantity": notional / strike * sign,
                "unit": asset,
            }
        )

    # Otherwise receive the notional back
    events.append(
        {
            "track": "",
            "time": maturity,
            "op": "+",
            "quantity": notional * sign,
            "unit": ccy,
        }
    )
    return timetable_from_dicts(events)


if __name__ == "__main__":
    # Create the rainbow option
    timetable = rainbow_timetable(
        "USD",
        ["SPX", "FTSE", "N225"],
        [5087, 7684, 39100],
        100_000,
        0.5,
        True,
    )
    print(timetable["events"].to_pandas())



"""
This module contains examples of fixed rate bonds.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List

import numpy as np
import pandas as pd
import pyarrow as pa

from qablet_contracts.ir.dcf import dcf_30_360 as dcf
from qablet_contracts.timetable import TS_EVENT_SCHEMA


def _const_dict_array(n, val):
    """Create a dictionary array of length n with constant values."""
    return pa.DictionaryArray.from_arrays(
        indices=np.full(n, 0, dtype=np.int64),
        dictionary=[val],
    )


@dataclass
class FixedCashFlows:
    """A set of **Fixed Cashflows** in a single currency. This example also shows how to create a timetable from arrays
    instead of a list of dictionaries, which is more efficient.

    Args:
        ccy: the currency of cashflows.
        times: a list or ndarray of cashflows times.
        amounts: a list ndarray of cashflows amounts.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = FixedCashFlows(
            "USD",
            [
                datetime(2023, 12, 31),
                datetime(2024, 6, 30),
                datetime(2024, 12, 31),
            ],
            [0.05, 0.05, 1.05],
        ).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time op  quantity unit
        0       2023-12-31 00:00:00+00:00  +      0.05  USD
        1       2024-06-30 00:00:00+00:00  +      0.05  USD
        2       2024-12-31 00:00:00+00:00  +      1.05  USD
    """

    ccy: str
    dates: List[datetime]
    amounts: List[float]
    track: str = ""

    def timetable(self):
        n = len(self.dates)
        return {
            "events": pa.RecordBatch.from_arrays(
                [
                    _const_dict_array(n, self.track),  # tracks
                    pa.array(self.dates),
                    _const_dict_array(n, "+"),  # ops
                    pa.array(self.amounts),
                    _const_dict_array(n, self.ccy),  # units
                ],
                schema=TS_EVENT_SCHEMA,
            ),
            "expressions": {},
        }


@dataclass
class FixedBond:
    """A **Fixed Rate Bond** pays a fixed rate at regular intervals, and the principal at maturity.

    Args:
        ccy: the currency of cashflows.
        coupon: the coupon rate per year.
        maturity: the maturity of the bond.
        freq: the number of coupon payments per year.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = FixedBond(
            "USD", 0.05, datetime(2023, 12, 31), datetime(2025, 12, 31), "2QE"
        ).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time op  quantity unit
        0       2024-06-30 00:00:00+00:00  +     0.025  USD
        1       2024-12-31 00:00:00+00:00  +     0.025  USD
        2       2025-06-30 00:00:00+00:00  +     0.025  USD
        3       2025-12-31 00:00:00+00:00  +     1.025  USD
    """

    ccy: str
    coupon: float
    accrual_start: datetime
    maturity: datetime
    freq: str = "2BQE"
    track: str = ""

    def timetable(self):
        # Coupon period dates including the start of first period, and end of last period.
        cpn_dates = pd.bdate_range(
            self.accrual_start,
            self.maturity,
            freq=self.freq,
            inclusive="both",
        )

        amounts = [
            dcf(end, start) * self.coupon
            for start, end in zip(cpn_dates[:-1], cpn_dates[1:])
        ]

        amounts[-1] += 1  # The last payment includes the principal
        return FixedCashFlows(
            self.ccy, cpn_dates[1:], amounts, self.track
        ).timetable()


if __name__ == "__main__":
    # Create a timetable from cashflows
    timetable = FixedCashFlows(
        "USD",
        [
            datetime(2023, 12, 31),
            datetime(2024, 6, 30),
            datetime(2024, 12, 31),
        ],
        [0.05, 0.05, 1.05],
    ).timetable()
    print("cf:\n", timetable["events"].to_pandas())

    # Create a fixed bond timetable
    timetable = FixedBond(
        "USD", 0.05, datetime(2023, 12, 31), datetime(2025, 12, 31), "2QE"
    ).timetable()
    print("bond:\n", timetable["events"].to_pandas())



from datetime import datetime

import pandas as pd

from qablet_contracts.eq.autocall import AutoCallable
from qablet_contracts.eq.barrier import OptionKO
from qablet_contracts.eq.cliquet import Accumulator
from qablet_contracts.eq.forward import ForwardOption


def test_classes():
    # Autocallable
    start = datetime(2024, 3, 31)
    maturity = datetime(2024, 9, 30)
    barrier_dates = pd.date_range(
        start, maturity, freq="ME", inclusive="right"
    )
    tt = AutoCallable(
        "USD", "AAPL", 100, 80, start, maturity, 102, barrier_dates, 0.092
    ).timetable()
    assert len(tt["events"]) == 7

    # Barrier
    start = datetime(2024, 3, 31)
    maturity = datetime(2024, 9, 30)
    barrier_dates = pd.date_range(
        start, maturity, freq="ME", inclusive="right"
    )
    tt = OptionKO(
        "USD", "EQ", 100, maturity, True, 102, "Up/Out", barrier_dates
    ).timetable()
    assert len(tt["events"]) == 9

    # Cliquet
    fix_dates = pd.bdate_range(
        datetime(2021, 12, 31), datetime(2024, 12, 31), freq="2BQE"
    )
    global_floor = 0.0
    local_floor = -0.03
    local_cap = 0.05
    tt = Accumulator(
        "USD", "SPX", fix_dates, global_floor, local_floor, local_cap
    ).timetable()
    assert len(tt["events"]) == 9

    # ForwardOption
    tt = ForwardOption(
        "USD",
        "SPX",
        1.0,
        datetime(2024, 3, 31),
        datetime(2024, 9, 30),
        True,
    ).timetable()
    assert len(tt["events"]) == 4



"""
This module contains examples of creating timetables for equity vanilla options.
These methods creates a timetable with floating point for time, which is now being deprecated.
See qablet_contracts/eq/vanilla to create timetables with timestamps.
"""

from qablet_contracts.timetable import timetable_from_dicts


def _option_events(ccy, asset_name, strike, maturity, is_call, track):
    events = [
        {
            "track": track,
            "time": maturity,
            "op": ">",
            "quantity": 0,
            "unit": ccy,
        },
        {
            "track": track,
            "time": maturity,
            "op": "+",
            "quantity": -strike if is_call else strike,
            "unit": ccy,
        },
        {
            "track": track,
            "time": maturity,
            "op": "+",
            "quantity": 1 if is_call else -1,
            "unit": asset_name,
        },
    ]
    return events


def option_timetable(
    ccy: str,
    asset_name: str,
    strike: float,
    maturity: float,
    is_call: bool,
    track: str = "",
) -> dict:
    """Create timetable for an **Equity Vanilla Option**.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        strike: the option strike.
        maturity: the maturity of the option in years.
        is_call: true if the option is a call.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = option_timetable("USD", "AAPL", 190, 1.0, True)  # Call
        >>> tt["events"].to_pandas()
          track  time op  quantity  unit
        0         1.0  >       0.0   USD
        1         1.0  +    -190.0   USD
        2         1.0  +       1.0  AAPL

        >>> tt = option_timetable("USD", "AAPL", 190, 1.0, False)  # Put
        >>> tt["events"].to_pandas()
          track  time op  quantity  unit
        0         1.0  >       0.0   USD
        1         1.0  +     190.0   USD
        2         1.0  +      -1.0  AAPL
    """

    events = _option_events(ccy, asset_name, strike, maturity, is_call, track)

    return timetable_from_dicts(events)


if __name__ == "__main__":
    # Create the option
    timetable = option_timetable("USD", "SPX", 2900, 0.5, True, track="ID_01")
    print(timetable["events"].to_pandas())



"""
This module contains examples of equity cliquet contracts.
"""

from dataclasses import dataclass, field
from datetime import datetime
from typing import List

import numpy as np
import pandas as pd

from qablet_contracts.timetable import EventsMixin


@dataclass
class Accumulator(EventsMixin):
    """In an **Accumulator** the returns over consecutive periods are

    - subject to a local floor and cap
    - accumulated by adding
    - the accumulated payoff is subject to a global floor

    Args:
        ccy: the currency of the bond.
        asset_name: the name of the underlying asset.
        fix_dates: the fixing times of the cliquet.
        global_floor: the global floor of the cliquet.
        local_floor: the local floor of the cliquet.
        local_cap: the local cap of the cliquet.
        track: an optional identifier for the contract.

    Examples:
        >>> fix_dates = pd.bdate_range(datetime(2021, 12, 31), datetime(2024, 12, 31), freq="2BQE")
        >>> tt = Accumulator("USD", "SPX", fix_dates, 0.0, -0.03, 0.05).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time   op  quantity     unit
        0   NaN 2021-12-31 00:00:00+00:00  NaN       0.0     INIT
        1   NaN 2022-06-30 00:00:00+00:00  NaN       0.0  CALCFIX
        2   NaN 2022-12-30 00:00:00+00:00  NaN       0.0  CALCFIX
        3   NaN 2023-06-30 00:00:00+00:00  NaN       0.0  CALCFIX
        4   NaN 2023-12-29 00:00:00+00:00  NaN       0.0  CALCFIX
        5   NaN 2024-06-28 00:00:00+00:00  NaN       0.0  CALCFIX
        6   NaN 2024-12-31 00:00:00+00:00  NaN       0.0  CALCFIX
        7       2024-12-31 00:00:00+00:00    >       0.0      USD
        8       2024-12-31 00:00:00+00:00    +       1.0      ACC
    """

    ccy: str
    asset_name: str
    fix_dates: List[datetime]
    global_floor: float
    local_floor: float
    local_cap: float
    track: str = ""
    state: dict = field(default_factory=dict)

    def events(self):
        maturity = self.fix_dates[-1]

        events = [
            {
                "track": None,
                "time": self.fix_dates[0],
                "op": None,
                "quantity": 0,
                "unit": "INIT",  # initialize accumulator
            }
        ]
        for fixing_time in self.fix_dates[1:]:
            events.append(
                {
                    "track": None,
                    "time": fixing_time,
                    "op": None,
                    "quantity": 0,
                    "unit": "CALCFIX",  # update accumulator
                }
            )
        events.append(
            {
                "track": self.track,
                "time": maturity,
                "op": ">",  # global floor
                "quantity": self.global_floor,
                "unit": self.ccy,
            }
        )
        events.append(
            {
                "track": self.track,
                "time": maturity,
                "op": "+",  # pay the accumulated amount
                "quantity": 1,
                "unit": "ACC",
            }
        )
        return events

    def expressions(self):
        last_acc = self.state.get("ACC", 0.0)

        if "S_PREV" in self.state:

            def accumulator_init_fn(inputs):
                [s] = inputs
                return [last_acc, self.state["S_PREV"]]  # [ACC, S_PREV]
        else:

            def accumulator_init_fn(inputs):
                [s] = inputs
                return [last_acc, s]  # [ACC, S_PREV]

        def accumulator_update_fn(inputs):
            [s, s_prev, a] = inputs

            ret = s / s_prev - 1.0  # ret = S / S_PREV - 1
            ret = np.maximum(self.local_floor, ret)
            ret = np.minimum(self.local_cap, ret)

            return [a + ret, s]  # [ACC, S_PREV]

        return {
            "INIT": {
                "type": "snapper",
                "inp": [self.asset_name],
                "fn": accumulator_init_fn,
                "out": ["ACC", "S_PREV"],
            },
            "CALCFIX": {
                "type": "snapper",
                "inp": [self.asset_name, "S_PREV", "ACC"],
                "fn": accumulator_update_fn,
                "out": ["ACC", "S_PREV"],
            },
        }


if __name__ == "__main__":
    # Create the cliquet
    fix_dates = pd.bdate_range(
        datetime(2021, 12, 31), datetime(2024, 12, 31), freq="2BQE"
    )

    global_floor = 0.0
    local_floor = -0.03
    local_cap = 0.05
    timetable = Accumulator(
        "USD",
        "SPX",
        fix_dates,
        global_floor,
        local_floor,
        local_cap,
        state={"S_PREV": 1.0},
    ).timetable()

    print(timetable["events"].to_pandas())



"""
This module contains examples of day count fraction calculations.
"""

from datetime import datetime, timedelta


def _is_eom(dt):
    return (dt + timedelta(days=1)).month != dt.month


def dcf_30_360(end: datetime, start: datetime) -> float:
    """Calculate US 30/360 daycount fraction. It is a simple
    implementation, and may not capture all the edge cases.

    Args:
        end: the end date.
        start: the start date.
    """
    if _is_eom(start):
        d1 = 30
    else:
        d1 = start.day

    if d1 == 30 and end.day == 31:
        d2 = 30
    else:
        d2 = end.day

    return (
        (end.year - start.year)
        + (end.month - start.month) / 12
        + (d2 - d1) / 360
    )



"""
This module contains examples of zero coupon bond and related contracts.
These methods creates a timetable with floating point for time, which is now being deprecated.
See qablet_contracts/bnd/zero to create timetables with timestamps.
"""

from qablet_contracts.timetable import timetable_from_dicts


def zcb_timetable(ccy: str, maturity: float, track: str = "") -> dict:
    """A **zero coupon bond** pays a single fixed payment at a future time.

    Args:
        ccy: the currency of the bond.
        maturity: the maturity of the bond in years.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = zcb_timetable("USD", 1)
        >>> tt["events"].to_pandas()
          track  time op  quantity unit
        0         1.0  +       1.0  USD
    """
    events = [
        {
            "track": track,
            "time": maturity,
            "op": "+",
            "quantity": 1,
            "unit": ccy,
        }  # get bond notional at bond expiration
    ]
    return timetable_from_dicts(events)


def zbp_timetable(
    ccy: str,
    opt_maturity: float,
    bond_maturity: float,
    strike: float,
    track: str = "",
) -> dict:
    """A **zero coupon bond put** offers the holder the option to sell a zero coupon bond for
    a fixed strike price, on the option maturity date.

    Args:
        ccy: the currency of the bond.
        opt_maturity: the maturity of the option in years.
        bond_maturity: the maturity of the bond in years.
        strike: the option strike.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = zbp_timetable("USD", 0.5, 1.0, 0.95)
        >>> tt["events"].to_pandas()
          track  time op  quantity unit
        0         0.5  >      0.00  USD
        1         0.5  +      0.95  USD
        2         1.0  +     -1.00  USD
    """

    events = [
        {
            "track": track,
            "time": opt_maturity,
            "op": ">",
            "quantity": 0,
            "unit": ccy,
        },  # Choose greater of nothing (get 0) or exercise (continue to remaining events)
        {
            "track": track,
            "time": opt_maturity,
            "op": "+",
            "quantity": strike,
            "unit": ccy,
        },  # get strike at expiration
        {
            "track": track,
            "time": bond_maturity,
            "op": "+",
            "quantity": -1,
            "unit": ccy,
        },  # pay bond notional at bond expiration
    ]
    return timetable_from_dicts(events)


if __name__ == "__main__":
    # Create a zero coupon bond timetable
    timetable = zcb_timetable("USD", 1)
    print("zcb:\n", timetable["events"].to_pandas())

    timetable = zbp_timetable("USD", 0.5, 1.0, 0.95)
    print("zbp:\n", timetable["events"].to_pandas())



"""
This module contains examples of autocallable notes.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List

import numpy as np
import pandas as pd

from qablet_contracts.ir.dcf import dcf_30_360 as dcf
from qablet_contracts.timetable import EventsMixin


@dataclass
class AutoCallable(EventsMixin):
    """In an **Autocallable Note** the note is called if the asset price is above the barrier level
    on any of the barrier observation dates. On being called the note pays the principal and the coupon accreted
    till the call date. At maturity, if the note is not called by then,

    - if the asset is above strike, it pays the principal and the coupon at maturity.
    - if the asset is below strike, it pays the notional scaled down.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        initial_spot: the initial spot price of the asset.
        strike: downside participation below this strike.
        maturity: the maturity of the option in years.
        barrier: the note is called above this barrier level.
        barrier_dates: the barrier observation points.
        cpn_rate: the coupon rate.
        track: an optional identifier for the contract.

    Examples:
        >>> start = datetime(2024, 3, 31)
        >>> maturity = datetime(2024, 9, 30)
        >>> barrier_dates = pd.date_range(start, maturity, freq="ME", inclusive="right")
        >>> tt = AutoCallable(
            "USD", "AAPL", 100, 80, start, maturity, 102, barrier_dates, 0.092
        ).timetable()
        >>> print(tt["events"].to_pandas())
          track                      time    op    quantity    unit
        0       2024-04-30 00:00:00+00:00  CALL  100.769613     USD
        1       2024-05-31 00:00:00+00:00  CALL  101.545149     USD
        2       2024-06-30 00:00:00+00:00  CALL  102.326654     USD
        3       2024-07-31 00:00:00+00:00  CALL  103.114173     USD
        4       2024-08-31 00:00:00+00:00  CALL  103.907753     USD
        5       2024-09-30 00:00:00+00:00  CALL  104.707441     USD
        6       2024-09-30 00:00:00+00:00     +    1.000000  PAYOFF
    """

    ccy: str
    asset_name: str
    initial_spot: float
    strike: float
    accrual_start: datetime
    maturity: datetime
    barrier: float
    barrier_dates: List[datetime]
    cpn_rate: float
    track: str = ""

    def events(self):
        events = []
        # Autocall events
        for barrier_date in self.barrier_dates:
            # daycount_fraction
            frac = dcf(barrier_date, self.accrual_start)
            events.append(
                {
                    "track": self.track,
                    "time": barrier_date,
                    "op": "CALL",
                    "quantity": 100 * np.exp(frac * self.cpn_rate),
                    "unit": self.ccy,
                }
            )

        # payoff at maturity
        events.append(
            {
                "track": "",
                "time": self.maturity,
                "op": "+",
                "quantity": 1.0,
                "unit": "PAYOFF",
            }
        )
        return events

    def expressions(self):
        # Define the autocall condition
        def ko_fn(inputs):
            [S] = inputs
            return [S > (self.barrier * self.initial_spot / 100)]

        call = {
            "type": "phrase",
            "inp": [self.asset_name],
            "fn": ko_fn,
        }

        # Define the final payoff
        fixed_pay = 100 * np.exp(
            dcf(self.maturity, self.accrual_start) * self.cpn_rate
        )

        def payoff_fn(inputs):
            [s] = inputs
            eq_pay = s * (100 / self.initial_spot)
            return [np.where(eq_pay < self.strike, eq_pay, fixed_pay)]

        payoff = {
            "type": "phrase",
            "inp": [self.asset_name],
            "fn": payoff_fn,
        }

        return {"PAYOFF": payoff, "CALL": call}


if __name__ == "__main__":
    # Create the option
    start = datetime(2024, 3, 31)
    maturity = datetime(2024, 9, 30)
    barrier_dates = pd.date_range(
        start, maturity, freq="ME", inclusive="right"
    )
    timetable = AutoCallable(
        "USD", "AAPL", 100, 80, start, maturity, 102, barrier_dates, 0.092
    ).timetable()
    print(timetable["events"].to_pandas())



from qablet_contracts.timetable import EVENT_SCHEMA, TS_EVENT_SCHEMA


def test_old_schema():
    assert EVENT_SCHEMA.names == ["track", "time", "op", "quantity", "unit"]
    assert TS_EVENT_SCHEMA.names == ["track", "time", "op", "quantity", "unit"]



"""
Create an autocallable note timetable.
These methods creates a timetable with floating point for time, which is now being deprecated.
See qablet_contracts/eq/autocall to create timetables with timestamps.
"""

import numpy as np
import pyarrow as pa

from qablet_contracts.timetable import EVENT_SCHEMA


def autocallable_timetable(
    ccy: str,
    asset_name: str,
    initial_spot: float,
    strike: float,
    maturity: float,
    barrier: float,
    barrier_pts: int,
    cpn_rate: float,
    track: str = "",
) -> dict:
    """Create timetable for an **Autocallable Note**.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        initial_spot: the initial spot price of the asset.
        strike: downside participation below this strike.
        maturity: the maturity of the option in years.
        barrier: the note is called above this barrier level.
        barrier_pts: the number of barrier observation points. If 0, then the note is not autocallable.
        cpn_rate: the coupon rate.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = autocallable_timetable("USD", "AAPL", 100, 80, 1, 102, 4, 0.092)
        >>> tt["events"].to_pandas()
          track  time    op    quantity     unit
        0        0.25  CALL  102.326654      USD
        1        0.50  CALL  104.707441      USD
        2        0.75  CALL  107.143621      USD
        3        1.00  CALL  109.636482      USD
        4        1.00     +    1.000000   PAYOFF
    """

    events_list = []
    # Autocall events
    if barrier_pts:
        for barrier_time in np.linspace(0, maturity, barrier_pts + 1)[1:]:
            events_list.append(
                {
                    "track": track,
                    "time": barrier_time,
                    "op": "CALL",
                    "quantity": 100 * np.exp(barrier_time * cpn_rate),
                    "unit": ccy,
                }
            )

    # payoff at maturity
    events_list.append(
        {
            "track": "",
            "time": maturity,
            "op": "+",
            "quantity": 1.0,
            "unit": "PAYOFF",
        }
    )

    # Define the autocall condition
    def ko_fn(inputs):
        [S] = inputs
        return [S > (barrier * initial_spot / 100)]

    call = {
        "type": "phrase",
        "inp": [asset_name],
        "fn": ko_fn,
    }

    # Define the final payoff
    fixed_pay = 100 * np.exp(maturity * cpn_rate)

    def payoff_fn(inputs):
        [s] = inputs
        eq_pay = s * (100 / initial_spot)
        return [np.where(eq_pay < strike, eq_pay, fixed_pay)]

    calcpay = {
        "type": "phrase",
        "inp": [asset_name],
        "fn": payoff_fn,
    }

    events_table = pa.RecordBatch.from_pylist(events_list, schema=EVENT_SCHEMA)
    return {
        "events": events_table,
        "expressions": {"PAYOFF": calcpay, "CALL": call},
    }


if __name__ == "__main__":
    # Create the option
    timetable = autocallable_timetable(
        "USD", "AAPL", 100, 80, 1, 102, 4, 0.092
    )
    print(timetable["events"].to_pandas())



"""
Utils for creating barrier options timetable.
These methods creates a timetable with floating point for time, which is now being deprecated.
See qablet_contracts/eq/barrier to create timetables with timestamps.
"""

import numpy as np
import pyarrow as pa

from qablet_contracts.equity.vanilla import _option_events
from qablet_contracts.timetable import EVENT_SCHEMA


def ko_option_timetable(
    ccy: str,
    asset_name: str,
    strike: float,
    maturity: float,
    is_call: bool,
    barrier: float,
    barrier_type: str,
    barrier_pts: int,
    rebate: float = 0,
    track: str = "",
) -> dict:
    """Create timetable for a **Knock Out Option**.

    Args:
        ccy: the currency of the option.
        asset_name: the name of the underlying asset.
        strike: the option strike.
        maturity: the maturity of the option in years.
        is_call: true if the option is a call.
        barrier: the barrier level.
        barrier_type: the type of barrier option, e.g "Dn/Out" or "Up/Out".
        barrier_pts: the number of barrier observation points.
        rebate: the rebate amount.
        track: an optional identifier for the contract.

    Examples:
        >>> tt = ko_option_timetable("USD", "EQ", 100, 0.2, True, 100, "Dn/Out", 5)
        >>> tt["events"].to_pandas()
          track  time  op  quantity unit
        0        0.00  KO       0.0  USD
        1        0.04  KO       0.0  USD
        2        0.08  KO       0.0  USD
        3        0.12  KO       0.0  USD
        4        0.16  KO       0.0  USD
        5        0.20  KO       0.0  USD
        6        0.20   >       0.0  USD
        7        0.20   +    -100.0  USD
        8        0.20   +       1.0   EQ
    """

    events = []
    for barrier_time in np.linspace(0, maturity, barrier_pts + 1):
        events.append(
            {
                "track": track,
                "time": barrier_time,
                "op": "KO",
                "quantity": rebate,
                "unit": ccy,
            }
        )

    vanilla_events = _option_events(
        ccy, asset_name, strike, maturity, is_call, track
    )

    events.extend(vanilla_events)

    # Define the knockout expression (KO)
    if barrier_type == "Dn/Out":

        def ko_fn(inputs):
            [S] = inputs
            return [S < barrier]
    elif barrier_type == "Up/Out":

        def ko_fn(inputs):
            [S] = inputs
            return [S > barrier]
    else:
        raise ValueError(f"Unknown barrier type: {barrier_type}")

    ko = {
        "type": "phrase",
        "inp": [asset_name],
        "fn": ko_fn,
    }

    events_table = pa.RecordBatch.from_pylist(events, schema=EVENT_SCHEMA)
    return {"events": events_table, "expressions": {"KO": ko}}


if __name__ == "__main__":
    # Create the option
    timetable = ko_option_timetable(
        "USD", "EQ", 100, 0.2, True, 100, "Dn/Out", 5
    )
    print(timetable["events"].to_pandas())
